// Libraries
import { Injectable } from '@angular/core';
import { Actions, Effect, ofType } from '@ngrx/effects';
import { of } from 'rxjs';
import {
  map,
  flatMap,
  withLatestFrom,
  filter,
  catchError,
} from 'rxjs/operators';
import { GlobalState } from '@app/core/state';
import { Store } from '@ngrx/store';

// Services
import { DegreePlannerApiService } from '@app/degree-planner/services/api.service';
import { getDegreePlannerState } from '@app/degree-planner/store/selectors';

// Actions
import {
  AddCourse,
  AddCourseSuccess,
  AddSaveForLater,
  AddSaveForLaterSuccess,
  CourseActionTypes,
  CourseError,
  MoveCourseBetweenTerms,
  MoveCourseBetweenTermsSuccess,
  RemoveCourse,
  RemoveCourseSuccess,
  RemoveSaveForLater,
  RemoveSaveForLaterSuccess,
} from '@app/degree-planner/store/actions/course.actions';

// Models
import { DegreePlan } from '@app/core/models/degree-plan';
import { Course, CourseBase } from '@app/core/models/course';

@Injectable()
export class CourseEffects {
  constructor(
    private actions$: Actions,
    private api: DegreePlannerApiService,
    private store$: Store<GlobalState>,
  ) {}

  @Effect()
  MoveCourseBetweenTerms$ = this.actions$.pipe(
    ofType<MoveCourseBetweenTerms>(CourseActionTypes.MoveCourseBetweenTerms),

    withLatestFrom(this.store$.select(getDegreePlannerState)),
    filter(([_, state]) => typeof state.visibleDegreePlan !== undefined),

    // Get term data for the degree plan specified by the roadmap ID.
    flatMap(([action, state]) => {
      const roadmapId = (state.visibleDegreePlan as DegreePlan).roadmapId;
      const recordId = action.payload.id;
      const termCode = action.payload.to;
      return this.api
        .updateCourseTerm(roadmapId, recordId, termCode)
        .pipe(map(() => action));
    }),

    map(action => new MoveCourseBetweenTermsSuccess(action.payload)),

    catchError(error => {
      return of(
        new CourseError({
          message: 'Error moving course',
          error,
        }),
      );
    }),
  );

  @Effect()
  AddCourse$ = this.actions$.pipe(
    ofType<AddCourse>(CourseActionTypes.AddCourse),

    withLatestFrom(this.store$.select(getDegreePlannerState)),
    filter(([_, state]) => state.visibleDegreePlan !== undefined),

    // Get term data for the degree plan specified by the roadmap ID.
    flatMap(([action, state]) => {
      // TODO error handle the API calls
      const roadmapId = (state.visibleDegreePlan as DegreePlan).roadmapId;
      const { subjectCode, termCode, courseId } = action.payload;

      const addCourse$ = this.api.addCourse(
        roadmapId,
        subjectCode,
        courseId,
        termCode,
      );

      const courseBaseToCourse$ = addCourse$.pipe(
        map<CourseBase, Course>(courseBase => ({
          ...courseBase,
          subject: state.subjects[courseBase.subjectCode],
        })),
      );

      const toSuccessAction$ = courseBaseToCourse$.pipe(
        map(course => new AddCourseSuccess({ course })),
      );

      return toSuccessAction$;
    }),

    catchError(error => {
      return of(
        new CourseError({
          message: 'Error adding course',
          error,
        }),
      );
    }),
  );

  @Effect()
  RemoveCourse$ = this.actions$.pipe(
    ofType<RemoveCourse>(CourseActionTypes.RemoveCourse),

    withLatestFrom(this.store$.select(getDegreePlannerState)),
    filter(([_, state]) => state.visibleDegreePlan !== undefined),

    // Get term data for the degree plan specified by the roadmap ID.
    flatMap(([action, state]) => {
      const roadmapId = (state.visibleDegreePlan as DegreePlan).roadmapId;
      const recordId = action.payload.recordId;

      const removeCourse$ = this.api.removeCourse(roadmapId, recordId);

      const toSuccessAction$ = removeCourse$.pipe(
        map(() => new RemoveCourseSuccess({ recordId })),
      );

      return toSuccessAction$;
    }),

    catchError(error => {
      return of(
        new CourseError({
          message: 'Error removing course',
          error,
        }),
      );
    }),
  );

  @Effect()
  RemoveSavedForLater$ = this.actions$.pipe(
    ofType<RemoveSaveForLater>(CourseActionTypes.RemoveSaveForLater),

    flatMap(action => {
      const { subjectCode, courseId } = action.payload;
      return this.api
        .removeSavedForLater(subjectCode, courseId)
        .pipe(map(() => action));
    }),

    map(action => new RemoveSaveForLaterSuccess(action.payload)),

    catchError(error => {
      return of(
        new CourseError({
          message: 'Error removing saved-for-later course',
          error,
        }),
      );
    }),
  );

  @Effect()
  SaveForLater$ = this.actions$.pipe(
    ofType<AddSaveForLater>(CourseActionTypes.AddSaveForLater),

    flatMap(action => {
      const { subjectCode, courseId } = action.payload;
      return this.api
        .saveForLater(subjectCode, courseId)
        .pipe(map(() => action));
    }),

    map(action => new AddSaveForLaterSuccess(action.payload)),

    catchError(error => {
      return of(
        new CourseError({
          message: 'Error saving course for later',
          error,
        }),
      );
    }),
  );
}
